#include <config.h>
#include <ESP32Servo.h>
#include <AccelStepper.h>
#include <MultiStepper.h>
#include <string.h>
#include "BluetoothSerial.h"

#define DEBUG

// DEFINE GRBL SERIAL PORT
BluetoothSerial SerialBT;

// DEFINE STEPPER PINS
AccelStepper stepper1(AccelStepper::DRIVER, MOTOR_1_STEP_PIN, MOTOR_1_DIRECTION_PIN);
AccelStepper stepper2(AccelStepper::DRIVER, MOTOR_2_STEP_PIN, MOTOR_2_DIRECTION_PIN);
AccelStepper stepper3(AccelStepper::DRIVER, MOTOR_3_STEP_PIN, MOTOR_3_DIRECTION_PIN);
MultiStepper steppers;

// initializing servo
Servo penServo;

#define MOTOR_1_MAX_SPEED 3000
#define MOTOR_2_MAX_SPEED 3000
#define MOTOR_3_MAX_SPEED 3000

#define MOTOR_1_MAX_ACCELERATION 1000
#define MOTOR_2_MAX_ACCELERATION 1000
#define MOTOR_3_MAX_ACCELERATION 1000

#define LINE_BUFFER_LENGTH 1024

#define STEPPERS_DISABLED 1
#define STEPPERS_ENABLED 0

struct point
{
  float x;
  float y;
  float z;
  int speed;
};

struct point actuatorPos;

float StepInc = 1;
int StepDelay = 0;
int LineDelay = 50;

// float StepsPerMillimeterX = 0.86;
// float StepsPerMillimeterY = 0.5;

float StepsPerMillimeterX = 86.0;
float StepsPerMillimeterY = 50.0;

float Xmin = 0;
float Ymin = 0;

float Xpos = Xmin;
float Ypos = Ymin;



boolean verbose = false;
long positions[2];



float xPosition = 0.0;
float yPosition = 0.0;

void mov(long x, long y)
{
  // Convert desired position changes to scaled velocities
  float Vx = x * SCALING_FACTOR;
  float Vy = y * SCALING_FACTOR;
  // float Vz = z * SCALING_FACTOR;

  // Initialize velocities
  float v1 = 0;
  float v2 = 0;
  float v3 = 0;

  // Calculate wheel velocities for X direction (forward/backward)
  if (Vx != 0)
  {
    v2 = Vx / WHEEL_RADIUS;
    v3 = -Vx / WHEEL_RADIUS;
  }

  // Calculate wheel velocities for Y direction (sideways)
  if (Vy != 0)
  {
    v1 = 2 * Vy / WHEEL_RADIUS;
    v2 = Vy / WHEEL_RADIUS;
    v3 = Vy / WHEEL_RADIUS;
  }

  // Convert velocities to steps (using a time interval of 1 second for simplicity)
  positions[0] = v1 * STEPS_PER_REV / (2 * PI);
  positions[1] = v2 * STEPS_PER_REV / (2 * PI);
  positions[2] = v3 * STEPS_PER_REV / (2 * PI);

  steppers.moveTo(positions);

 
  steppers.runSpeedToPosition();

  stepper1.setCurrentPosition(0);
  stepper2.setCurrentPosition(0);
  stepper3.setCurrentPosition(0);
}

// void drawLine(float x1, float y1)
// {
//   float dx = x1 - xPosition;
//   float dy = y1 - yPosition;

// #ifdef DEBUG

//   // Debugging output for target positions
//   Serial.print("Target X: ");
//   Serial.println(x1);
//   Serial.print("Target Y: ");
//   Serial.println(y1);

//   // Debugging output for relative movements
//   Serial.print("Move X by: ");
//   Serial.println(dx);
//   Serial.print("Move Y by: ");
//   Serial.println(dy);
// #endif

//   // Update current positions based on movement
//   xPosition = x1;
//   yPosition = y1;

// #ifdef DEBUG

//   // Debugging output for new positions
//   Serial.print("xPosition: ");
//   Serial.println(xPosition);
//   Serial.print("yPosition: ");
//   Serial.println(yPosition);
// #endif

//   mov(dx, dy); // Call the actual move function with relative distances
// }

void drawLine(float x1, float y1) {

  if (verbose)
  {
    Serial.print("fx1, fy1: ");
    Serial.print(x1);
    Serial.print(",");
    Serial.print(y1);
    Serial.println("");
  }  

  

  if (verbose)
  {
    Serial.print("Xpos, Ypos: ");
    Serial.print(Xpos);
    Serial.print(",");
    Serial.print(Ypos);
    Serial.println("");
  }

  if (verbose)
  {
    Serial.print("x1, y1: ");
    Serial.print(x1);
    Serial.print(",");
    Serial.print(y1);
    Serial.println("");
  }

  //  Convert coordinates to steps
  x1 = (int)(x1*StepsPerMillimeterX);
  y1 = (int)(y1*StepsPerMillimeterY);
  float x0 = Xpos;
  float y0 = Ypos;

  //  Let's find out the change for the coordinates
  long dx = abs(x1-x0);
  long dy = abs(y1-y0);
  int sx = x0<x1 ? StepInc : -StepInc;
  int sy = y0<y1 ? StepInc : -StepInc;

  long i;
  long over = 0;

  if (dx > dy) {
    for (i=0; i<dx; ++i) {
      // myStepperX.step(sx);
      mov(sx, 0);
      over+=dy;
      if (over>=dx) {
        over-=dx;
        // myStepperY.step(sy);
        mov(0, sy);
      }
      delay(StepDelay);
    }
  }
  else {
    for (i=0; i<dy; ++i) {
      // myStepperY.step(sy);
      mov(0, sy);
      over+=dx;
      if (over>=dy) {
        over-=dy;
        // myStepperX.step(sx);
        mov(sx, 0);
      }
      delay(StepDelay);
    }    
  }

  if (verbose)
  {
    Serial.print("dx, dy:");
    Serial.print(dx);
    Serial.print(",");
    Serial.print(dy);
    Serial.println("");
  }

  if (verbose)
  {
    Serial.print("Going to (");
    Serial.print(x0);
    Serial.print(",");
    Serial.print(y0);
    Serial.println(")");
  }

  
  delay(LineDelay);
  
  Xpos = x1;
  Ypos = y1;
}

void processIncomingLine(char *line, int charNB)
{
  int currentIndex = 0;
  char buffer[64];
  struct point newPos;

  newPos.x = 0.0;
  newPos.y = 0.0;
  newPos.z = 0.0;
  newPos.speed = 0;

  int G_command = 0;

  char *indexX;
  char *indexY;
  char *indexZ;
  char *indexSpeed;

  while (currentIndex < charNB)
  {
    switch (line[currentIndex++])
    {
    case 'G':
      buffer[0] = line[currentIndex++];
      buffer[1] = line[currentIndex++];
      buffer[2] = '\0';
      G_command = atoi(buffer);
      switch (atoi(buffer))
      {
      case 00:
        // Serial.println(atoi(buffer));
      case 01:
        // Serial.println("G01");
        Serial.println('Function :: ');
        indexX = strchr(line + currentIndex++, 'X');
        indexY = strchr(line + currentIndex++, 'Y');
        indexZ = strchr(line + currentIndex++, 'Z');
        indexSpeed = strchr(line + currentIndex++, 'F');

        if (indexSpeed > 0)
        {
          newPos.speed = atof(indexSpeed + 1);
        }

        if (G_command == 1)
        {
          Serial.print("Speed :: ");
          Serial.println(newPos.speed);
          stepper1.setMaxSpeed(newPos.speed);
          stepper2.setMaxSpeed(newPos.speed);
          stepper3.setMaxSpeed((newPos.speed) / 2);
        }
        else if (G_command == 0)
        {

          stepper1.setMaxSpeed(MOTOR_1_MAX_SPEED);
          stepper2.setMaxSpeed(MOTOR_2_MAX_SPEED);
          stepper3.setMaxSpeed(MOTOR_3_MAX_SPEED);
        }

        if (indexY <= 0 && indexZ <= 0)
        { // If Y and Z are missing
          newPos.y = actuatorPos.y;
          newPos.z = actuatorPos.z;
          newPos.x = atof(indexX + 1);
        }
        else if (indexX <= 0 && indexZ <= 0)
        { // If X and Z are missing
          newPos.x = actuatorPos.x;
          newPos.z = actuatorPos.z;
          newPos.y = atof(indexY + 1);
        }
        else if (indexX <= 0 && indexY <= 0)
        { // If X and Y are missing
          newPos.x = actuatorPos.x;
          newPos.y = actuatorPos.y;
          newPos.z = atof(indexZ + 1);
        }
        else if (indexZ <= 0)
        { // If Z is missing
          newPos.x = atof(indexX + 1);
          newPos.y = atof(indexY + 1);
          newPos.z = actuatorPos.z;
        }
        else if (indexY <= 0)
        { // If Y is missing
          newPos.x = atof(indexX + 1);
          newPos.z = atof(indexZ + 1);
          newPos.y = actuatorPos.y;
        }
        else if (indexX <= 0)
        { // If X is missing
          newPos.y = atof(indexY + 1);
          newPos.z = atof(indexZ + 1);
          newPos.x = actuatorPos.x;
        }
        else
        { // All coordinates are present
          newPos.x = atof(indexX + 1);
          newPos.y = atof(indexY + 1);
          newPos.z = atof(indexZ + 1);
        }

        drawLine(newPos.x, newPos.y);
        actuatorPos.x = newPos.x;
        actuatorPos.y = newPos.y;
        actuatorPos.z = newPos.z;
        break;

      case 28:
        Ypos = 0.0;
        Xpos = 0.0;
        xPosition = 0.0;
        yPosition = 0.0;
        SerialBT.println("Home Position Set X = 0.0 Y = 0.0");
        // home();
        break;
      default:
        SerialBT.println("Unknown command");
        break;
      }
      break;
    case 'M':
      buffer[0] = line[currentIndex++];
      buffer[1] = line[currentIndex++];
      buffer[2] = '\0';
      switch (atoi(buffer))
      {
      case 03:
      {
        char *indexS = strchr(line + currentIndex++, 'S');
        int servoAngle = atoi(indexS + 1);
        if (servoAngle <= 180 && servoAngle >= 0)
        {
          penServo.write(servoAngle);
          delay(15);
#ifdef DEBUG
          Serial.println(servoAngle);
#endif
        }
      }
      break;

      default:
        SerialBT.print("Command not recognized : M");
        SerialBT.println(buffer);
        break;
      }
      break;
    default:
      SerialBT.print("Unknown command : ");
      SerialBT.println(buffer);
      break;
    }
  }
}

void processCommands()
{

  delay(200);
  char line[LINE_BUFFER_LENGTH];
  char c;
  int lineIndex;
  bool lineIsComment, lineSemiColon;

  lineIndex = 0;
  lineSemiColon = false;
  lineIsComment = false;

  while (1)
  {
    while (SerialBT.available() > 0)
    {
      c = SerialBT.read();
      if ((c == '\n') || (c == '\r'))
      {
        if (lineIndex > 0)
        {
          line[lineIndex] = '\0';
          if (verbose)
          {
            SerialBT.print("Received : ");
            SerialBT.println(line);
          }
          if (strcmp(line, "G28") == 0)
          {
            // home();
            Xpos = 0.0;
            Ypos = 0.0;
            xPosition = 0.0;
            yPosition = 0.0;
            SerialBT.println("Home Position Set X = 0.0 Y = 0.0");
          }
          else
          {
            processIncomingLine(line, lineIndex);
          }
          lineIndex = 0;
        }
        else
        {
        }
        lineIsComment = false;
        lineSemiColon = false;
        SerialBT.println("OK");
      }
      else
      {
        if ((lineIsComment) || (lineSemiColon))
        {
          if (c == ')')
            lineIsComment = false;
        }
        else
        {
          if (c <= ' ')
          {
          }
          else if (c == '/')
          {
          }
          else if (c == '(')
          {
            lineIsComment = true;
          }
          else if (c == ';')
          {
            lineSemiColon = true;
          }
          else if (lineIndex >= LINE_BUFFER_LENGTH - 1)
          {
            Serial.println("ERROR - lineBuffer overflow");
            lineIsComment = false;
            lineSemiColon = false;
          }
          else if (c >= 'a' && c <= 'z')
          {
            line[lineIndex++] = c - 'a' + 'A';
          }
          else
          {
            line[lineIndex++] = c;
          }
        }
      }
    }
  }
}

void setup()
{
  Serial.begin(115200);
  SerialBT.begin(MACHINE_NAME);

  pinMode(STEPPERS_DISABLE_PIN, OUTPUT);
  pinMode(STATUS_LED_PIN, OUTPUT);

  penServo.attach(SERVO_PIN);
  delay(15); // wait for 15 ms

  penServo.write(50);
  delay(15); // wait for 15 ms


  // Configure each stepper
  stepper1.setMaxSpeed(MOTOR_1_MAX_SPEED);
  stepper1.setAcceleration(MOTOR_1_MAX_ACCELERATION);
  stepper1.setCurrentPosition(0);
  stepper2.setMaxSpeed(MOTOR_2_MAX_SPEED);
  stepper2.setAcceleration(MOTOR_2_MAX_ACCELERATION);
  stepper2.setCurrentPosition(0);
  stepper3.setMaxSpeed(MOTOR_3_MAX_SPEED);
  stepper3.setAcceleration(MOTOR_3_MAX_ACCELERATION);
  stepper3.setCurrentPosition(0);
  steppers.addStepper(stepper1);
  steppers.addStepper(stepper2);
  steppers.addStepper(stepper3);

  Serial.println(" Poombata is Ready pair with bluetooth");
  // Serial.print("X range is from ");
  // Serial.print(Xmin);
  // Serial.print(" to ");
  // Serial.print(Xmax);
  // Serial.println(" mm.");
  // Serial.print("Y range is from ");
  // Serial.print(Ymin);
  // Serial.print(" to ");
  // Serial.print(Ymax);
  // Serial.println(" mm.");
  // Serial.print("Z range is from ");
  // Serial.print(Zmin);
  // Serial.print(" to ");
  // Serial.print(Zmax);
  // Serial.println(" mm.");

  digitalWrite(STEPPERS_DISABLE_PIN, STEPPERS_ENABLED);
}

void loop()
{

  processCommands();
}